#include "opencv2/opencv.hpp"

#ifdef _OPENMP
#include <omp.h>
#define NUM_THREADS_OPENMP num_threads(4)
#endif

#include "stattimer.hpp"
STimerList st(STimerList::reporterTSV, "stat.txt");

void putDeltaMilliSec(cv::Mat& img, cv::Point org,
                      double tick_begin, double tick_end)
{
    double delta = 1000.0 * (tick_end - tick_begin) / cv::getTickFrequency();
    cv::putText(img, cv::format("%.3f [ms]", delta), org,
                CV_FONT_HERSHEY_PLAIN, 1.0, CV_RGB(0, 255, 0));
}
void putMode(cv::Mat& img, cv::Point org, int mode)
{
    char *label[] = { "c: OpenCV function",
                      "n: naive implementation",
                      "o: optimized implementation"};
    for (int i = 0; i < sizeof(label)/sizeof(char *); i++) {
        cv::putText(img, (i == mode)? "*" : " ",
                    org + cv::Point(0, 12 * i),
                    CV_FONT_HERSHEY_PLAIN, 1.0, CV_RGB(0, 255, 0));
        cv::putText(img, label[i], org + cv::Point(15, 12 * i),
                    CV_FONT_HERSHEY_PLAIN, 1.0,
                    CV_RGB(0, (i == mode)? 255: 200, 0));
    }
}

void highlightFrameDiff(const cv::Mat& input_color,
                        cv::Mat& gray, 
                        cv::Mat& prev_gray,
                        cv::Mat& output_color)
{
    st.start("baseline OpenCV");

    // convert color
    cv::cvtColor(input_color, gray, CV_BGR2GRAY);

    // frame diff
    cv::Mat diff;
    cv::absdiff(gray, prev_gray, diff);
    cv::compare(diff, cv::Scalar(30), diff, cv::CMP_GT);

    std::vector<cv::Mat> channel(3);
    cv::split(input_color, channel);
    channel[0] = gray | diff;
    channel[1] = gray & ~diff;
    channel[2] = gray & ~diff;
    cv::merge(channel, output_color);

    // save current frame
    gray.copyTo(prev_gray);

    st.stop("baseline OpenCV");
}


void highlightFrameDiffNaive(const cv::Mat& input_color,
                             cv::Mat& gray, 
                             cv::Mat& prev_gray,
                             cv::Mat& output_color)
{
    st.start("baseline naive");
    
    const int width = input_color.cols;
    const int height = input_color.rows;

    // convert color
    for (int j = 0; j < height; j++) {
        for (int i = 0; i < width; i++) {
            cv::Vec3b p = input_color.at<cv::Vec3b>(j, i);
            uchar b = p[0];
            uchar g = p[1];
            uchar r = p[2];
            gray.at<uchar>(j, i) = (r * 306 + g * 601 + b * 117) / 1024;
        }
    }

    // frame diff
    for (int j = 0; j < height; j++) {
        for (int i = 0; i < width; i++) {
            uchar g = gray.at<uchar>(j, i);
            int diff =  g - prev_gray.at<uchar>(j, i);
            if (diff > 30 || diff < -30) {
                output_color.at<cv::Vec3b>(j, i)[0] = 255;
                output_color.at<cv::Vec3b>(j, i)[1] = 0;
                output_color.at<cv::Vec3b>(j, i)[2] = 0;
            } else {
                output_color.at<cv::Vec3b>(j, i)[0] = g;
                output_color.at<cv::Vec3b>(j, i)[1] = g;
                output_color.at<cv::Vec3b>(j, i)[2] = g;
            }
        }
    }

    // save current frame
    for (int j = 0; j < height; j++) {
        for (int i = 0; i < width; i++) {
            prev_gray.at<uchar>(j, i) = gray.at<uchar>(j, i);
        }
    }

    st.stop("baseline naive");
}

void highlightFrameDiffOptim(const cv::Mat& input_color,
                             cv::Mat& gray, 
                             cv::Mat& prev_gray,
                             cv::Mat& output_color)
{
    st.start("optimized");
    
    const int width = input_color.cols;
    const int height = input_color.rows;

    // convert color
#ifdef _OPENMP
#pragma omp parallel for NUM_THREADS_OPENMP
#endif
    for (int j = 0; j < height - 1; j++) {
        const cv::Vec3b *in_ptr = input_color.ptr<cv::Vec3b>(j);
        uchar *g_ptr = gray.ptr<uchar>(j);
        uchar tmpbuf[16];
        for (int i = 0; i < width; i += 4) {

            __m128i data = _mm_loadu_si128((__m128i *)(in_ptr + i));
            // BGRBGRBGRBGR(BGRB)

            __m128i bmask = _mm_setr_epi8(0,-1,-1,-1,
                                          3,-1,-1,-1,
                                          6,-1,-1,-1,
                                          9,-1,-1,-1);
            __m128i b4 = _mm_shuffle_epi8(data, bmask);
            __m128i bcoeff = _mm_setr_epi32(306, 306, 306, 306);
            b4 = _mm_mullo_epi32(b4, bcoeff);

            __m128i gmask = _mm_setr_epi8( 1,-1,-1,-1,
                                           4,-1,-1,-1,
                                           7,-1,-1,-1,
                                           10,-1,-1,-1);
            __m128i g4 = _mm_shuffle_epi8(data, gmask);
            __m128i gcoeff = _mm_setr_epi32(601, 601, 601, 601);
            g4 = _mm_mullo_epi32(g4, gcoeff);

            __m128i rmask = _mm_setr_epi8( 2,-1,-1,-1,
                                           5,-1,-1,-1,
                                           8,-1,-1,-1,
                                           11,-1,-1,-1);
            __m128i r4 = _mm_shuffle_epi8(data, rmask);
            __m128i rcoeff = _mm_setr_epi32(117, 117, 117, 117);
            r4 = _mm_mullo_epi32(r4, rcoeff);

            __m128i grayval = _mm_add_epi32(b4, g4);
            grayval = _mm_add_epi32(grayval, r4);
            grayval = _mm_srli_epi32(grayval, 10);

            _mm_store_si128((__m128i *)tmpbuf, grayval);
            g_ptr[i + 0] = tmpbuf[0];
            g_ptr[i + 1] = tmpbuf[4];
            g_ptr[i + 2] = tmpbuf[8];
            g_ptr[i + 3] = tmpbuf[12];
        }
    }
    { /* To avoid buffer overrun, the last line is not processed by SSE
       * (because we read 16 bytes while we process 4 pixels (= 12 bytes))
       */
        const cv::Vec3b *in_ptr = input_color.ptr<cv::Vec3b>(height - 1);
        uchar *g_ptr = gray.ptr<uchar>(height - 1);
        for (int i = 0; i < width; i++) {
            uchar b = in_ptr[i][0];
            uchar g = in_ptr[i][1];
            uchar r = in_ptr[i][2];
            g_ptr[i] = (r * 306 + g * 601 + b * 117) / 1024;
        }
    }

    // frame diff
#ifdef _OPENMP
#pragma omp parallel for NUM_THREADS_OPENMP
#endif
    for (int j = 0; j < height; j++) {
        uchar *g_ptr = gray.ptr<uchar>(j);
        uchar *prev_g_ptr = prev_gray.ptr<uchar>(j);
        cv::Vec3b *out_ptr = output_color.ptr<cv::Vec3b>(j);

        for (int i = 0; i < width; i += 4) {
            // 0th pixel
            uchar grayval = g_ptr[i];
            int diff =  grayval - prev_g_ptr[i];
            if (diff > 30 || diff < -30) {
                out_ptr[i][0] = 255;
                out_ptr[i][1] = 0;
                out_ptr[i][2] = 0;
            } else {
                out_ptr[i][0] = grayval;
                out_ptr[i][1] = grayval;
                out_ptr[i][2] = grayval;
            }

            // 1st pixel
            grayval = g_ptr[i + 1];
            diff =  grayval - prev_g_ptr[i + 1];
            if (diff > 30 || diff < -30) {
                out_ptr[i + 1][0] = 255;
                out_ptr[i + 1][1] = 0;
                out_ptr[i + 1][2] = 0;
            } else {
                out_ptr[i + 1][0] = grayval;
                out_ptr[i + 1][1] = grayval;
                out_ptr[i + 1][2] = grayval;
            }

            // 2nd pixel
            grayval = g_ptr[i + 2];
            diff =  grayval - prev_g_ptr[i + 2];
            if (diff > 30 || diff < -30) {
                out_ptr[i + 2][0] = 255;
                out_ptr[i + 2][1] = 0;
                out_ptr[i + 2][2] = 0;
            } else {
                out_ptr[i + 2][0] = grayval;
                out_ptr[i + 2][1] = grayval;
                out_ptr[i + 2][2] = grayval;
            }

            // 3rd pixel
            grayval = g_ptr[i + 3];
            diff =  grayval - prev_g_ptr[i + 3];
            if (diff > 30 || diff < -30) {
                out_ptr[i + 3][0] = 255;
                out_ptr[i + 3][1] = 0;
                out_ptr[i + 3][2] = 0;
            } else {
                out_ptr[i + 3][0] = grayval;
                out_ptr[i + 3][1] = grayval;
                out_ptr[i + 3][2] = grayval;
            }
        }
    }

    // save current frame
    memcpy(prev_gray.data, gray.data, width * height);

    st.stop("optimized");
}


int main()
{
    enum { OPENCV_FUNC, NAIVE, OPTIM } mode;
    mode = OPENCV_FUNC;

    cv::namedWindow("display", cv::WINDOW_AUTOSIZE);
    cv::VideoCapture cap(0);
    cv::Mat input, display;
    cap >> input;
    cv::Mat gray(input.rows, input.cols, CV_8U);
    cv::Mat prev_gray(input.rows, input.cols, CV_8U);
    cv::Mat output(input.rows, input.cols, CV_8UC3);

    while (1) {
        cap >> input;

        double t_begin = (double)cv::getTickCount();
        if (mode == OPENCV_FUNC) {
            highlightFrameDiff(input, gray, prev_gray, output);            
        } else if (mode == NAIVE) {
            highlightFrameDiffNaive(input, gray, prev_gray, output);
        } else if (mode == OPTIM) {
            highlightFrameDiffOptim(input, gray, prev_gray, output);
        }
        double t_end = (double)cv::getTickCount();

        output.copyTo(display);
        putDeltaMilliSec(display, cv::Point(20, 20), t_begin, t_end);
        putMode(display, cv::Point(20, 40), (int)mode);

        cv::imshow("display", display);
        int key;
        if ((key = cv::waitKey(30)) > 0) {
            if (key == 'q' || key == 0x1b) {
                break;
            } else if (key == 'c') {
                mode = OPENCV_FUNC;
            } else if (key == 'n') {
                mode = NAIVE;
            } else if (key == 'o') {
                mode = OPTIM;
            }
        }
    }

}
