#include "opencv2/opencv.hpp"

#ifdef _OPENMP
#include <omp.h>
#define NUM_THREADS_OPENMP num_threads(4)
#endif

#include <emmintrin.h>

void putDeltaMilliSec(cv::Mat& img, cv::Point org,
                      double tick_begin, double tick_end)
{
    double delta = 1000.0 * (tick_end - tick_begin) / cv::getTickFrequency();
    cv::putText(img, cv::format("%.3f [ms]", delta), org,
                CV_FONT_HERSHEY_PLAIN, 1.0, CV_RGB(0, 255, 0));
}

void putKernelSize(cv::Mat& img, cv::Point org, int ksize)
{
    cv::putText(img, cv::format("%d x %d kernel", ksize, ksize), org,
                CV_FONT_HERSHEY_PLAIN, 1.0, CV_RGB(0, 255, 0));
}

void putMode(cv::Mat& img, cv::Point org, int mode)
{
    char *label[] = { "c: OpenCV function",
                      "t: naive 2D convolusion",
                      "s: separable filter",
                      "o: separable filter; optimized"};
    for (int i = 0; i < sizeof(label)/sizeof(char *); i++) {
        cv::putText(img, (i == mode)? "*" : " ",
                    org + cv::Point(0, 12 * i),
                    CV_FONT_HERSHEY_PLAIN, 1.0, CV_RGB(0, 255, 0));
        cv::putText(img, label[i], org + cv::Point(15, 12 * i),
                    CV_FONT_HERSHEY_PLAIN, 1.0,
                    CV_RGB(0, (i == mode)? 255: 200, 0));
    }
}

int getKsize(double sigma)
{
    // from cv::createGaussianFilter() impl
    int ksize = cvRound(sigma * 3 * 2 + 1) | 1;
    return ksize;
}

void Gaussian2D(const cv::Mat& input, cv::Mat& output, double sigma)
{
    int ksize = getKsize(sigma);
    int margin = ksize / 2;
    cv::Mat w0 = cv::getGaussianKernel(ksize, sigma, CV_32F);

    cv::Mat w(ksize, ksize, CV_32F);
    for (int j = 0; j < ksize; j++) {
        for (int i = 0; i < ksize; i++) {
            w.at<float>(j, i) = w0.at<float>(j) * w0.at<float>(i);
        }
    }
    cv::Mat inputf(input.rows, input.cols, CV_32F);
    for (int j = margin; j < input.rows - margin; j++) {
        for (int i = margin; i < input.cols - margin; i++) {
            inputf.at<float>(j, i) = input.at<uchar>(j, i);
        }
    }

    output = 0;
    for (int j = margin; j < input.rows - margin; j++) {
        for (int i = margin; i < input.cols - margin; i++) {
            float sum = 0.0f;
            for (int n = -margin; n <= margin; n++) {
                for (int m = -margin; m <= margin; m++) {
                    sum += w.at<float>(n + margin, m + margin)
                        * inputf.at<float>(j + n, i + m);
                }
            }
            output.at<uchar>(j, i) = (uchar)sum;
        }
    }

}

void GaussianSepXY(const cv::Mat& input, cv::Mat& output, double sigma)
{
    int ksize = getKsize(sigma);
    int margin = ksize / 2;
    cv::Mat w = cv::getGaussianKernel(ksize, sigma, CV_32F);

    cv::Mat tmp(input.rows, input.cols, CV_32F);
    cv::Mat inputf(input.rows, input.cols, CV_32F);
    for (int j = margin; j < input.rows - margin; j++) {
        for (int i = margin; i < input.cols - margin; i++) {
            inputf.at<float>(j, i) = input.at<uchar>(j, i);
        }
    }

    for (int j = 0; j < input.rows; j++) {
        for (int i = margin; i < input.cols - margin; i++) {
            float sum = 0.0f;
            for (int m = -margin; m <= margin; m++) {
                sum += w.at<float>(m + margin) * inputf.at<float>(j, i + m);
            }
            tmp.at<float>(j, i) = sum;
        }
    }

    output = 0;
    for (int j = margin; j < input.rows - margin; j++) {
        for (int i = 0; i < input.cols; i++) {
            float sum = 0.0f;
            for (int m = -margin; m <= margin; m++) {
                sum += w.at<float>(m + margin) * tmp.at<float>(j + m, i);
            }
            output.at<uchar>(j, i) = (uchar)sum;
        }
    }
}

void GaussianSepOptim(const cv::Mat& input, cv::Mat& output, double sigma)
{
    int ksize = getKsize(sigma);
    int margin = ksize / 2;
    cv::Mat w = cv::getGaussianKernel(ksize, sigma, CV_32F);
    w.convertTo(w, CV_32S, 256);
    int *wp = w.ptr<int>(0);

    int height = input.rows;
    int width = input.cols;
    int width8_ = (width / 8) * 8;
    int width16_ = (width / 16) * 16;
    int width_2margin8_ = ((width - 2 * margin) / 8) * 8;
    int width_2margin16_ = ((width - 2 * margin) / 16) * 16;

    cv::Mat input32s(height, width, CV_32S);
    cv::Mat tmp(height, width, CV_32S);


#ifdef _OPENMP
#pragma omp parallel for NUM_THREADS_OPENMP
#endif
    for (int j = 0; j < height; j++) {
        const uchar *inp = input.ptr<uchar>(j);
        int *inp32sp = input32s.ptr<int>(j);
        for (int i = 0; i < width16_; i += 16) {
            __m128i zero = _mm_setzero_si128();
            __m128i pix8u = _mm_loadu_si128((__m128i *)(inp + i));
            __m128i pix16u_lo = _mm_unpacklo_epi8(pix8u, zero);
            __m128i pix16u_hi = _mm_unpackhi_epi8(pix8u, zero);

            __m128i pix32s0 = _mm_unpacklo_epi16(pix16u_lo, zero);
            __m128i pix32s1 = _mm_unpackhi_epi16(pix16u_lo, zero);
            __m128i pix32s2 = _mm_unpacklo_epi16(pix16u_hi, zero);
            __m128i pix32s3 = _mm_unpackhi_epi16(pix16u_hi, zero);

            _mm_storeu_si128((__m128i *)(inp32sp + i), pix32s0);
            _mm_storeu_si128((__m128i *)(inp32sp + 4 + i), pix32s1);
            _mm_storeu_si128((__m128i *)(inp32sp + 8 + i), pix32s2);
            _mm_storeu_si128((__m128i *)(inp32sp + 12 + i), pix32s3);

        }
        for (int i = width16_; i < width; i++) {
            inp32sp[i] = (int)inp[i];
        }
    }

#ifdef _OPENMP
#pragma omp parallel for NUM_THREADS_OPENMP
#endif
    for (int j = 0; j < height; j++) {
        int *inp32sp= input32s.ptr<int>(j);
        int *tmpp = tmp.ptr<int>(j);

        for (int i = margin; i < margin + width_2margin16_; i += 16) {
            __m128i sum0 = _mm_setzero_si128();
            __m128i sum1 = sum0;
            __m128i sum2 = sum0;
            __m128i sum3 = sum0;
            for (int m = -margin; m <= margin; m++) {
                int w = wp[m + margin];
                __m128i weight = _mm_cvtsi32_si128(w);
                weight = _mm_shuffle_epi32(weight, 0);
                __m128i pix32s0 = _mm_loadu_si128((__m128i *)(inp32sp + i + m));
                __m128i pix32s1 = _mm_loadu_si128((__m128i *)(inp32sp + i + 4 + m));
                __m128i pix32s2 = _mm_loadu_si128((__m128i *)(inp32sp + i + 8 + m));
                __m128i pix32s3 = _mm_loadu_si128((__m128i *)(inp32sp + i + 12 + m));
                sum0 = _mm_add_epi32(sum0, _mm_mullo_epi32(weight, pix32s0));
                sum1 = _mm_add_epi32(sum1, _mm_mullo_epi32(weight, pix32s1));
                sum2 = _mm_add_epi32(sum2, _mm_mullo_epi32(weight, pix32s2));
                sum3 = _mm_add_epi32(sum3, _mm_mullo_epi32(weight, pix32s3));
            }
            _mm_storeu_si128((__m128i *)(tmpp + i), sum0);
            _mm_storeu_si128((__m128i *)(tmpp + 4 + i), sum1);
            _mm_storeu_si128((__m128i *)(tmpp + 8 + i), sum2);
            _mm_storeu_si128((__m128i *)(tmpp + 12 + i), sum3);
        }
        for (int i = margin + width_2margin16_; i < width - margin; i++) {
            int sum = 0;
            for (int m = -margin; m <= margin; m++) {
                sum += wp[m + margin] * inp32sp[i + m];
            }
            tmpp[i] = sum;
        }
    }

    memset(output.data, 0, width * margin);

    uchar shf0_arr[16] = { 0, 4, 8, 12, 0x80, 0x80, 0x80, 0x80,
                           0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80};
    uchar shf1_arr[16] = { 0x80, 0x80, 0x80, 0x80, 0, 4, 8, 12, 
                           0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80};
    uchar shf2_arr[16] = { 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 
                           0, 4, 8, 12, 0x80, 0x80, 0x80, 0x80};
    uchar shf3_arr[16] = { 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 
                           0x80, 0x80, 0x80, 0x80, 0, 4, 8, 12 };
    __m128i shf0 = _mm_loadu_si128((__m128i *)shf0_arr);
    __m128i shf1 = _mm_loadu_si128((__m128i *)shf1_arr);
    __m128i shf2 = _mm_loadu_si128((__m128i *)shf2_arr);
    __m128i shf3 = _mm_loadu_si128((__m128i *)shf3_arr);

#ifdef _OPENMP
#pragma omp parallel for NUM_THREADS_OPENMP
#endif
    for (int j = margin; j < height - margin; j++) {
        int *tmpp = tmp.ptr<int>(j - margin);
        uchar *outp = output.ptr<uchar>(j);
        int step = (int)(tmp.step / sizeof(int));
        for (int i = 0; i < width16_; i += 16) {
            __m128i sum0 = _mm_setzero_si128();
            __m128i sum1 = sum0;
            __m128i sum2 = sum0;
            __m128i sum3 = sum0;
            int *pixp = tmpp + i;
            for (int m = -margin; m <= margin; m++, pixp += step) {
                int w = wp[m + margin];
                __m128i weight = _mm_cvtsi32_si128(w);
                weight = _mm_shuffle_epi32(weight, 0);
                __m128i pix32s0 = _mm_loadu_si128((__m128i *)pixp);
                __m128i pix32s1 = _mm_loadu_si128((__m128i *)(pixp + 4));
                __m128i pix32s2 = _mm_loadu_si128((__m128i *)(pixp + 8));
                __m128i pix32s3 = _mm_loadu_si128((__m128i *)(pixp + 12));
                sum0 = _mm_add_epi32(sum0, _mm_mullo_epi32(weight, pix32s0));
                sum1 = _mm_add_epi32(sum1, _mm_mullo_epi32(weight, pix32s1));
                sum2 = _mm_add_epi32(sum2, _mm_mullo_epi32(weight, pix32s2));
                sum3 = _mm_add_epi32(sum3, _mm_mullo_epi32(weight, pix32s3));
            }
            sum0 = _mm_srli_epi32(sum0, 16);
            sum1 = _mm_srli_epi32(sum1, 16);
            sum2 = _mm_srli_epi32(sum2, 16);
            sum3 = _mm_srli_epi32(sum3, 16);

            __m128i o0 = _mm_shuffle_epi8(sum0, shf0);
            __m128i o1 = _mm_shuffle_epi8(sum1, shf1);
            __m128i o2 = _mm_shuffle_epi8(sum2, shf2);
            __m128i o3 = _mm_shuffle_epi8(sum3, shf3);
            __m128i out = _mm_or_si128(o0, o1);
            out = _mm_or_si128(out, o2);
            out = _mm_or_si128(out, o3);
            _mm_storeu_si128((__m128i *)(outp + i), out);

        }
        for (int i = width16_; i < width; i++) {
            int sum = 0;
            int *pixp = tmp.ptr<int>(j - margin) + i;
            for (int m = -margin; m <= margin; m++, pixp += step) {
                sum += wp[m + margin] * (*pixp);
            }
            outp[i] = (uchar)(sum >> 16);
        }
    }

    memset(output.ptr<uchar>(height - margin), 0, width * margin);
}

int main()
{
    enum { OPENCV_FUNC, TWO_DIM, SEP_XY, SEP_OPTIM } mode;
    mode = OPENCV_FUNC;
    
    cv::namedWindow("display", cv::WINDOW_AUTOSIZE);
    cv::VideoCapture cap(0);
    cv::Mat input, display;
    cap >> input;
    cv::Mat output(input.rows, input.cols, CV_8U);

    int sigma = 3;
    cv::createTrackbar("sigma", "display", &sigma, 100);


    while (1) {
        cap >> input;
        cv::cvtColor(input, input, CV_BGR2GRAY);

	if (sigma < 1) { sigma = 1; }

        double t_begin = (double)cv::getTickCount();
        if (mode == OPENCV_FUNC) {
            cv::GaussianBlur(input, output, cv::Size(0, 0), (double)sigma);
        } else if (mode == TWO_DIM) {
            Gaussian2D(input, output, sigma);
        } else if (mode == SEP_XY) {
            GaussianSepXY(input, output, sigma);
        } else if (mode == SEP_OPTIM) {
            GaussianSepOptim(input, output, sigma);
        }
        double t_end = (double)cv::getTickCount();

        cv::cvtColor(output, display, CV_GRAY2BGR);
        putDeltaMilliSec(display, cv::Point(20, 20), t_begin, t_end);
        putKernelSize(display, cv::Point(20, 40), getKsize(sigma));
        putMode(display, cv::Point(20, 60), (int)mode);


        cv::imshow("display", display);
        int key;
        if ((key = cv::waitKey(30)) > 0) {
            if (key == 'q' || key == 0x1b) {
                break;
            } else if (key == 'c') { // OpenCV function
                mode = OPENCV_FUNC;
            } else if (key == 't') { // two dimensional
                mode = TWO_DIM;
            } else if (key == 's') { // separable
                mode = SEP_XY;
            } else if (key == 'o') { // separable optimized
                mode = SEP_OPTIM;
            }
        }
    }

    return 0;
}
